#include "custom_double_spin_box.hpp"

#include <QKeyEvent>
#include <QLineEdit>
#include <QtGlobal>

#include <algorithm>
#include <cmath>
#include <ranges>

CustomDoubleSpinBox::CustomDoubleSpinBox(QWidget *parent)
    : QDoubleSpinBox(parent) {}

void CustomDoubleSpinBox::stepBy(int steps) {
    if (steps == 0)
        return;

    auto newValue = value();

    // If the current value is no multiple of singleStep, we floor/ceil to the
    // nearest multiple
    if (!qFuzzyIsNull(std::fmod(value(), singleStep()))) {
        if (steps < 0) {
            newValue = std::floor(value() / singleStep()) * singleStep();
            ++steps;
        } else {
            newValue = std::ceil(value() / singleStep()) * singleStep();
            --steps;
        }
    }

    // For every other step, we move one whole step size
    newValue += steps * singleStep();
    // make sure we stay within min/max
    newValue = std::clamp(newValue, minimum(), maximum());

    setValue(newValue);
}

void CustomDoubleSpinBox::keyPressEvent(QKeyEvent *event) {

    auto isInt = false;
    event->text().toInt(&isInt);

    // Ignore input of non-ints, e.g. arrow keys.
    // Also don't do anything if there is already something selected.
    if (isInt && !lineEdit()->hasSelectedText()) {
        const auto lineEditText = lineEdit()->text();
        const auto cursorPosition = lineEdit()->cursorPosition();

        // Test if there is a decimal point to the left of the cursor.
        const auto decimalSepPosition = std::ranges::find(
            lineEditText | std::views::take(cursorPosition), '.');
        if (decimalSepPosition - std::ranges::begin(lineEditText) <
            cursorPosition) {

            // Only apply our tweak if there are the maximum number of decimal
            // numbers written. If there aren't, than the default behaviour of
            // inserting the new number and removing none works fine.
            const auto nrDecimals =
                std::ranges::end(lineEditText) - decimalSepPosition - 1;
            if (nrDecimals >= decimals()) {

                // We overwrite the character to our right by adding it to the
                // current (empty) selection.
                lineEdit()->setSelection(cursorPosition, 1);
            }
        }
    }

    // Proceed to handle the input like always, just with the potentially edited
    // selection.
    QDoubleSpinBox::keyPressEvent(event);
}
